/***************************** Include Files *********************************/
#include <stdio.h>
#include <stdarg.h>

#include <metal/log.h>
#include <metal/sys.h>

#include "xil_io.h"
#include "xil_printf.h"
#include "xparameters.h"
#include "xil_assert.h"
#include "xil_types.h"
#include "xstatus.h"
#include "xiicps.h"
#include "sleep.h"
#include "xrfdc.h"


/******************** Constant Definitions **********************************/
#define ENABLE_METAL_PRINTS
// RFDC defines
#define RFDC_DEVICE_ID 	XPAR_XRFDC_0_DEVICE_ID
#define RFDC_BASE       XPAR_XRFDC_0_BASEADDR
#define RFDC_DEV_NAME    XPAR_XRFDC_0_DEV_NAME

#define I2C_SLEEP_US 1000 /* I2C sleep period */
#define I2C_ADDR_BUS_SWITCH 0X74 /* Bus switch i2c address */
#define I2C_ADDR_I2C2SPI_BRIDGE 0X2F /* I2C2SPI bridge i2c address */
#define I2C_SWITCH_SELECT_I2C2SPI_BRIDGE (1 << 5) /* Switch value for bridge */
#define I2C_MUX_SEL_0 (1 << 1) /* MUX_SEL0 GPIO bit */
#define I2C_MUX_SEL_1 (1 << 2) /* MUX_SEL1 GPIO bit */


#define GPIO_DEVICE_ADDR	XPAR_AXI_GPIO_0_BASEADDR 	/* GPIO base address */
#define GPIO_DATA_REG 		0 							/* GPIO data register offset address */
#define GPIO_CONTROL_REG 	4 							/* GPIO control register offset address */

#define RFCLK_LMX2594_1 0 /* I0 on MUX and SS3 on Bridge */
#define RFCLK_LMX2594_2 1 /* I1 on MUX and SS2 on Bridge */
#define RFCLK_LMK 2 /* I2 on MUX and SS1 on Bridge */
#define RFCLK_CHIP_NUM 3
#define SELECT_SPI_SDO(X) (1 << (3 - X)) /* Value which routs MUX */

#define LMX_RESET_VAL 2 /* Reset value for LMX */
#define LMK_RESET_VAL 0X80 /* Reset value for LMK04828 */
#define LMX_ADDRESS_SHIFT 16 /* LMX address shift value */
#define LMX_ADDRESS_BITFIELD 0X7F /* LMX address bitfield */
#define LMK_ADDRESS_SHIFT 8 /* LMK address shift value */
#define LMK_ADDRESS_BITFIELD 0X1FFF /* LMK address bitfield */

#define LMK_COUNT 136
#define LMX2594_COUNT 116
/**************************** Type Definitions *******************************/

/***************** Macros (Inline Functions) Definitions *********************/

/************************** Function Prototypes ******************************/
void my_metal_default_log_handler(enum metal_log_level level,
			       const char *format, ...);
int CLK104_I2C_Init(void);
int I2C_RdData(XIicPs *Iic, u8 Addr, u8 *Val, u8 Len);
int I2C_WrData(XIicPs *Iic, u8 Addr, u8 *Val, u8 Len);
int CLK104_I2C_Switch_Sel(void);
u32 CLK104_GPIO_Init(void);
u32 CLK104_SPI_Init(void);
int CLK104_Init(void);

u32 CLK104_WriteReg(u32 ChipId, u32 data);
u32 CLK104_ResetChip(u32 ChipId);
u32 CLK104_LMK_Config();
u32 CLK104_LMX_Config(u32 ChipID);
int print_rfdc_status(XRFdc *InstancePtr);
/*********************** Device instance definitions *************************/
XIicPs clk104_i2c;
XRFdc RFdcInst;

/************************** Variable Definitions *****************************/
struct metal_device *deviceptr = NULL;

#ifdef __BAREMETAL__
metal_phys_addr_t metal_phys = RFDC_BASE;
static struct metal_device CustomDev = {
	/* RFdc device */
	.name = RFDC_DEV_NAME,
	.bus = NULL,
	.num_regions = 1,
	.regions = {
		{
			.virt = (void *)RFDC_BASE,
			.physmap = &metal_phys,
			.size = 0x40000,
			.page_shift = (unsigned)(-1),
			.page_mask = (unsigned)(-1),
			.mem_flags = 0x0,
			.ops = {NULL},
		}
	},
	.node = {NULL},
	.irq_num = 0,
	.irq_info = NULL,
};
#endif


u32 LMK_CKin[LMK_COUNT] = {
    // 250MHz
	0x000090, 0x000010, 0x000200, 0x000306, 0x0004D0, 0x00055B, 0x000600, 0x000C51, 0x000D04, 0x01006C, 0x010155, 0x010255, 0x010301, 0x010422, 0x010500, 0x010673, 0x010703, 0x01086C, 0x010955, 0x010A55,
	0x010B00, 0x010C22, 0x010D00, 0x010EF0, 0x010F30, 0x01106C, 0x011155, 0x011255, 0x011301, 0x011422, 0x011500, 0x011673, 0x011703, 0x01186C, 0x011955, 0x011A55, 0x011B01, 0x011C22, 0x011D00, 0x011E72,
	0x011F03, 0x012078, 0x012155, 0x012255, 0x012301, 0x012422, 0x012500, 0x012670, 0x012733, 0x01286C, 0x012955, 0x012A55, 0x012B00, 0x012C22, 0x012D00, 0x012EF0, 0x012F30, 0x01306C, 0x013155, 0x013255,
	0x013301, 0x013422, 0x013500, 0x013673, 0x013703, 0x013820, 0x013903, 0x013A01, 0x013B2C, 0x013C00, 0x013D01, 0x013E03, 0x013F02, 0x014009, 0x014100, 0x014200, 0x014331, 0x0144FF, 0x01457F, 0x014618,
	0x01471A, 0x014806, 0x014946, 0x014A06, 0x014B06, 0x014C00, 0x014D00, 0x014EC0, 0x014F7F, 0x015013, 0x015102, 0x015200, 0x015300, 0x01547D, 0x015500, 0x01567D, 0x015703, 0x0158C0, 0x015907, 0x015AD0,
	0x015BDA, 0x015C20, 0x015D00, 0x015E00, 0x015F3E, 0x016000, 0x016104, 0x016264, 0x016300, 0x016400, 0x0165A0, 0x0171AA, 0x017202, 0x017C15, 0x017D33, 0x016600, 0x016700, 0x016819, 0x016959, 0x016A20,
	0x016B00, 0x016C00, 0x016D00, 0x016E1B, 0x017300, 0x018200, 0x018300, 0x018400, 0x018500, 0x018800, 0x018900, 0x018A00, 0x018B00, 0x1FFD00, 0x1FFE00, 0x1FFF53,
};
u32 LMX2594_FREQ_250M_2000M[LMX2594_COUNT] = {
	0x000002, 0x000000,
	0x700000, 0x6F0000, 0x6E0000, 0x6D0000, 0x6C0000, 0x6B0000, 0x6A0000, 0x690021, 0x680000, 0x670000, 0x663F80,
	0x650011, 0x640000, 0x630000, 0x620200, 0x610888, 0x600000, 0x5F0000, 0x5E0000, 0x5D0000, 0x5C0000, 0x5B0000,
	0x5A0000, 0x590000, 0x580000, 0x570000, 0x560000, 0x55D300, 0x540001, 0x530000, 0x521E00, 0x510000, 0x506666,
	0x4F0026, 0x4E0001, 0x4D0000, 0x4C000C, 0x4B0840, 0x4A0000, 0x49003F, 0x480001, 0x470081, 0x46C350, 0x450000,
	0x4403E8, 0x430000, 0x4201F4, 0x410000, 0x401388, 0x3F0000, 0x3E0322, 0x3D00A8, 0x3C0000, 0x3B0001, 0x3A8001,
	0x390020, 0x380000, 0x370000, 0x360000, 0x350000, 0x340820, 0x330080, 0x320000, 0x314180, 0x300300, 0x2F0300,
	0x2E07FC, 0x2DC0DF, 0x2C1FA0, 0x2B0000, 0x2A0000, 0x290000, 0x280000, 0x270001, 0x260000, 0x250104, 0x240140,
	0x230004, 0x220000, 0x211E21, 0x200393, 0x1F43EC, 0x1E318C, 0x1D318C, 0x1C0488, 0x1B0002, 0x1A0DB0, 0x190624,
	0x18071A, 0x17007C, 0x160001, 0x150401, 0x14E048, 0x1327B7, 0x120064, 0x110064, 0x100080, 0x0F064F, 0x0E1E40,
	0x0D4000, 0x0C5001, 0x0B00A8, 0x0A10D8, 0x090604, 0x082000, 0x0740B2, 0x06C802, 0x0500C8, 0x040C43, 0x030642,
	0x020500, 0x010809, 0x00241C,
	0x00241C,
};
u32 LMX2594_FREQ_250M_6400M[LMX2594_COUNT] = {
	0x000002, 0x000000,
	0x700000, 0x6F0000, 0x6E0000, 0x6D0000, 0x6C0000, 0x6B0000, 0x6A0000, 0x690021, 0x680000, 0x670000, 0x663F80,
	0x650011, 0x640000, 0x630000, 0x620200, 0x610888, 0x600000, 0x5F0000, 0x5E0000, 0x5D0000, 0x5C0000, 0x5B0000,
	0x5A0000, 0x590000, 0x580000, 0x570000, 0x560000, 0x55D300, 0x540001, 0x530000, 0x521E00, 0x510000, 0x506666,
	0x4F0026, 0x4E0001, 0x4D0000, 0x4C000C, 0x4B0800, 0x4A0000, 0x49003F, 0x480001, 0x470081, 0x46C350, 0x450000,
	0x4403E8, 0x430000, 0x4201F4, 0x410000, 0x401388, 0x3F0000, 0x3E0322, 0x3D00A8, 0x3C0000, 0x3B0001, 0x3A8001,
	0x390020, 0x380000, 0x370000, 0x360000, 0x350000, 0x340820, 0x330080, 0x320000, 0x314180, 0x300300, 0x2F0300,
	0x2E07FC, 0x2DC0DF, 0x2C1FA0, 0x2B0000, 0x2A0000, 0x290000, 0x280000, 0x270001, 0x260000, 0x250204, 0x240200,
	0x230004, 0x220000, 0x211E21, 0x200393, 0x1F03EC, 0x1E318C, 0x1D318C, 0x1C0488, 0x1B0002, 0x1A0DB0, 0x190624,
	0x18071A, 0x17007C, 0x160001, 0x150401, 0x14E048, 0x1327B7, 0x120064, 0x110064, 0x100080, 0x0F064F, 0x0E1E40,
	0x0D4000, 0x0C5001, 0x0B00A8, 0x0A10D8, 0x090604, 0x082000, 0x0740B2, 0x06C802, 0x0500C8, 0x040C43, 0x030642,
	0x020500, 0x010809, 0x00241C,
	0x00241C,
};





int main()
{
	int Status;
	XRFdc_Config *ConfigPtr;
	XRFdc_IPStatus IPStatus;

	xil_printf("Hello RFSoC World!\n\r");

#ifdef ENABLE_METAL_PRINTS
    xil_printf("=== Metal log enabled ===\n\r");

	struct metal_init_params init_param = {	\
			.log_handler	= my_metal_default_log_handler,	\
			.log_level		= METAL_LOG_DEBUG,			\

	};
#else
	struct metal_init_params init_param = METAL_INIT_DEFAULTS;
#endif

	if (metal_init(&init_param)) {
		xil_printf("ERROR: Failed to run metal initialization\n");
		return XRFDC_FAILURE;
	}
	metal_log(METAL_LOG_DEBUG, "\n Test metal log in %s\r\n", __func__);


	xil_printf("Configuring the data converter clocks...\r\n");

	Status = CLK104_Init();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	Status = CLK104_ResetChip(RFCLK_LMX2594_1);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_ResetChip(RFCLK_LMX2594_2);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_ResetChip(RFCLK_LMK);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_LMK_Config();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_LMX_Config(RFCLK_LMX2594_1);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_LMX_Config(RFCLK_LMX2594_2);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}



	/* Initialize the RFdc driver. */
	ConfigPtr = XRFdc_LookupConfig(RFDC_DEVICE_ID);
	if (ConfigPtr == NULL) {
		xil_printf("Failed to init RFdc driver\r\n");
		return XRFDC_FAILURE;
	}else{
		xil_printf("\n\rDeviceID: %d \r\nSilicon Revision: %d\r\n", ConfigPtr->DeviceId, ConfigPtr->SiRevision);
	}

	/* Register & MAP RFDC to Libmetal */
#ifdef __BAREMETAL__
	deviceptr = &CustomDev;
#endif
	Status = XRFdc_RegisterMetal(&RFdcInst, RFDC_DEVICE_ID, &deviceptr);
	if (Status != XRFDC_SUCCESS) {
		return XRFDC_FAILURE;
	}
	/* Initializes the controller */
	Status = XRFdc_CfgInitialize(&RFdcInst, ConfigPtr);
	if (Status != XRFDC_SUCCESS) {
		xil_printf("Failed to init RFdc controller\r\n");
		return XRFDC_FAILURE;
	}

	print_rfdc_status(&RFdcInst);




	XRFdc_Distribution_Settings Distribution_Settings;
	memset(&Distribution_Settings, 0, sizeof(Distribution_Settings));
	// DAC Tile
	Distribution_Settings.SourceTileId = XRFDC_TILE_ID1;
	Distribution_Settings.SourceType = XRFDC_DAC_TILE;
	Distribution_Settings.EdgeTileIds[0] = XRFDC_TILE_ID3;
	Distribution_Settings.EdgeTileIds[1] = XRFDC_TILE_ID0;
	Distribution_Settings.EdgeTypes[0] = XRFDC_DAC_TILE;
	Distribution_Settings.EdgeTypes[1] = XRFDC_DAC_TILE;
	Distribution_Settings.DistRefClkFreq = 6400;
	Distribution_Settings.DistributedClock = XRFDC_DIST_OUT_RX;
	Distribution_Settings.SampleRates[1][0] = 6400;
	Distribution_Settings.SampleRates[1][1] = 6400;
	Distribution_Settings.SampleRates[1][2] = 6400;
	Distribution_Settings.SampleRates[1][3] = 6400;
	Status = XRFdc_SetClkDistribution(&RFdcInst, &Distribution_Settings);
//	if (Status != XRFDC_SUCCESS) {
//		return XRFDC_FAILURE;
//	}

	// ADC Tile
	Distribution_Settings.SourceTileId = XRFDC_TILE_ID1;
	Distribution_Settings.SourceType = XRFDC_ADC_TILE;
	Distribution_Settings.EdgeTileIds[0] = XRFDC_TILE_ID3;
	Distribution_Settings.EdgeTileIds[1] = XRFDC_TILE_ID0;
	Distribution_Settings.EdgeTypes[0] = XRFDC_ADC_TILE;
	Distribution_Settings.EdgeTypes[1] = XRFDC_ADC_TILE;
	Distribution_Settings.DistRefClkFreq = 2000;
	Distribution_Settings.DistributedClock = XRFDC_DIST_OUT_RX;
	Distribution_Settings.SampleRates[0][0] = 2000;
	Distribution_Settings.SampleRates[0][1] = 2000;
	Distribution_Settings.SampleRates[0][2] = 2000;
	Distribution_Settings.SampleRates[0][3] = 2000;

	Status = XRFdc_SetClkDistribution(&RFdcInst, &Distribution_Settings);
//	if (Status != XRFDC_SUCCESS) {
//		return XRFDC_FAILURE;
//	}

	XRFdc_Distribution_System_Settings DistributionArray;
	Status = XRFdc_GetClkDistribution(&RFdcInst, &DistributionArray);
	if (Status != XRFDC_SUCCESS) {
		return XRFDC_FAILURE;
	}

	print_rfdc_status(&RFdcInst);

    return 0;
}



int print_rfdc_status(XRFdc *InstancePtr){
	int Status, i;
	XRFdc_IPStatus IPStatus;
	Status = XRFdc_GetIPStatus(InstancePtr, &IPStatus);
	if (Status != XRFDC_SUCCESS) {
		return XRFDC_FAILURE;
	}

	xil_printf("======Report RFDC IP Status======\r\n");
	xil_printf("---------------------------------\r\n");
	xil_printf("DAC Tile Status:\r\n");
	for(i=0; i<4; i++){
		xil_printf("DAC Tile %d: ", i);
		if(IPStatus.DACTileStatus[i].IsEnabled != 1){
			xil_printf("Tile is not enabled \r\n");
			continue;
		}
		else if(IPStatus.DACTileStatus[i].PowerUpState != 1){
			xil_printf("Tile is not powered up\r\n");
			continue;
		}
		else if(IPStatus.DACTileStatus[i].TileState != 0xf){
			xil_printf("Tile State is not 0xf, stuck at 0x%x \r\n", IPStatus.DACTileStatus[i].TileState);
			continue;
		}
		else{
			xil_printf("Tile is ready, PLL State is %d\r\n", IPStatus.DACTileStatus[i].PLLState);
		}
	}
	xil_printf("---------------------------------\r\n");
	xil_printf("ADC Tile Status:\r\n");
	for(i=0; i<4; i++){
		xil_printf("ADC Tile %d: ", i);
		if(IPStatus.ADCTileStatus[i].IsEnabled != 1){
			xil_printf("Tile is not enabled \r\n");
			continue;
		}
		else if(IPStatus.ADCTileStatus[i].PowerUpState != 1){
			xil_printf("Tile is not powered up\r\n");
			continue;
		}
		else if(IPStatus.ADCTileStatus[i].TileState != 0xf){
			xil_printf("Tile State is not 0xF, stuck at 0x%x \r\n", IPStatus.ADCTileStatus[i].TileState);
			continue;
		}
		else{
			xil_printf("Tile is ready, PLL State is %d\r\n", IPStatus.ADCTileStatus[i].PLLState);
		}
	}
	xil_printf("======Report RFDC IP Status Done======\r\n");

	return XRFDC_SUCCESS;
}

int CLK104_I2C_Init(void){
	int Status;
	XIicPs_Config *i2c_inst;
	u32 ClkRate = 100000;

	/* i2c1 */
	i2c_inst = XIicPs_LookupConfig(XPAR_PSU_I2C_1_DEVICE_ID);
	if (NULL == i2c_inst) {
		return XST_FAILURE;
	}

	Status = XIicPs_CfgInitialize(&clk104_i2c, i2c_inst, i2c_inst->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	Status = XIicPs_SetSClk(&clk104_i2c, ClkRate);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	return XST_SUCCESS;
}

int I2C_RdData(XIicPs *Iic, u8 Addr, u8 *Val, u8 Len){
	int Status;
	Status = XIicPs_MasterRecvPolled(Iic, Val, Len, Addr);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	while (XIicPs_BusIsBusy(Iic))
		;
	usleep(I2C_SLEEP_US);
	return XST_SUCCESS;
}

int I2C_WrData(XIicPs *Iic, u8 Addr, u8 *Val, u8 Len){
	int Status;
	Status = XIicPs_MasterSendPolled(Iic, Val, Len, Addr);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	while (XIicPs_BusIsBusy(Iic))
		;
	usleep(I2C_SLEEP_US);
	return XST_SUCCESS;
}

int CLK104_I2C_Switch_Sel(void){
	u8 val = 0;
	int Status;

	/* Read I2C bus switch configuration */
	Status = I2C_RdData(&clk104_i2c, I2C_ADDR_BUS_SWITCH, &val, 1);
	if (Status == XST_FAILURE) {
		return XST_FAILURE;
	}

	/* Enable i2c2spi bridge */
	val |= I2C_SWITCH_SELECT_I2C2SPI_BRIDGE;
	/* Write new I2C bus switch configuration */
	Status = I2C_WrData(&clk104_i2c, I2C_ADDR_BUS_SWITCH, &val, 1);
	if (Status == XST_FAILURE) {
		return XST_FAILURE;
	}
	return XST_SUCCESS;
}

u32 CLK104_GPIO_Init(void){
	Xil_Out32(GPIO_DEVICE_ADDR + GPIO_CONTROL_REG, 0xfffffffC);
	return XST_SUCCESS;
}

u32 CLK104_SPI_Init(void){
	int Status;
	u8 tx[2] = { 0xf0, 0x03 };
	Status = I2C_WrData(&clk104_i2c, I2C_ADDR_I2C2SPI_BRIDGE, tx, 2);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	return XST_SUCCESS;
}
int CLK104_Init(void){
	int Status;
	Status = CLK104_I2C_Init();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_I2C_Switch_Sel();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_GPIO_Init();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	Status = CLK104_SPI_Init();
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}
	return XST_SUCCESS;
}



u32 CLK104_WriteReg(u32 ChipId, u32 data){
	Xil_AssertNonvoid(ChipId < RFCLK_CHIP_NUM);
	u8 tx[4] = { SELECT_SPI_SDO(ChipId), (data >> 16) & 0xff, (data >> 8) & 0xff,
		     data & 0xff };
	/* Write register */
	return I2C_WrData(&clk104_i2c, I2C_ADDR_I2C2SPI_BRIDGE, tx, 4);
}
u32 CLK104_ResetChip(u32 ChipId){
	Xil_AssertNonvoid(ChipId < RFCLK_CHIP_NUM);
	u32 val = LMX_RESET_VAL;
	if (ChipId == RFCLK_LMK)
		val = LMK_RESET_VAL;

	if (XST_SUCCESS != CLK104_WriteReg(ChipId, val)) {
		return XST_FAILURE;
	}
	if (XST_SUCCESS != CLK104_WriteReg(ChipId, 0)) {
		return XST_FAILURE;
	}
	return XST_SUCCESS;
}
u32 CLK104_LMK_Config(){
	for (int i = 0; i < LMK_COUNT; i++) {
		if (XST_SUCCESS !=
		    CLK104_WriteReg(RFCLK_LMK, LMK_CKin[i])) {
			return XST_FAILURE;
		}
	}
	return XST_SUCCESS;
}
u32 CLK104_LMX_Config(u32 ChipID){
	int Status;
	// ADC
	if(ChipID == RFCLK_LMX2594_1){
		for (int i = 0; i < LMX2594_COUNT; i++) {
			Status = CLK104_WriteReg(ChipID, LMX2594_FREQ_250M_2000M[i]);
			if (Status != XST_SUCCESS) {
			return XST_FAILURE;
			}
		}
	}
	// DAC
	else if(ChipID == RFCLK_LMX2594_2){
		for (int i = 0; i < LMX2594_COUNT; i++) {
			Status = CLK104_WriteReg(ChipID, LMX2594_FREQ_250M_6400M[i]);
			if (Status != XST_SUCCESS) {
			return XST_FAILURE;
			}
		}
	}
	else{
		return XST_FAILURE;
	}

	return XST_SUCCESS;
}


/*****************************************************************************/
/**
*
* My libmetal logger
* Intercepts log prints and adjusts \r\n prints to display the some on a uart
* or through a jtagUart.
*
******************************************************************************/
void my_metal_default_log_handler(enum metal_log_level level,
			       const char *format, ...)
{
	char msg[1024];
	char msgOut[1048];
	char *outPtr;
	int i;

	va_list args;
	static const char *level_strs[] = {
		"metal: emergency: ",
		"metal: alert:     ",
		"metal: critical:  ",
		"metal: error:     ",
		"metal: warning:   ",
		"metal: notice:    ",
		"metal: info:      ",
		"metal: debug:     ",
	};

	va_start(args, format);
	vsnprintf(msg, sizeof(msg), format, args);
	va_end(args);

	//replace single \n with \n\r
	outPtr = msgOut;
	for(i=0; i<1024; i++) {
		// if /n/r or /r/n combo
		if ((msg[i] == '\r' && msg[i+1] == '\n') ||
				(msg[i] == '\n' && msg[i+1] == '\r')) {
			*outPtr++ = msg[i++];
		} else if(msg[i] == '\n') {
			//if first char in string is \n, then remove
			if(i==0) {
				continue;
			} else {
				*outPtr++ = '\r';
			}
		}
		*outPtr++ = msg[i];
		if(msg[i] == 0) {
			break;
		}
	}
	//if line doesn't end with \n\r, then add it
	if( (msg[i-1] != '\n') && (msg[i-1] != '\r') ) {
		*(outPtr-1) = '\r';
		*outPtr++ = '\n';
		*outPtr++ = 0;
	}

	if (level <= METAL_LOG_EMERGENCY || level > METAL_LOG_DEBUG)
		level = METAL_LOG_EMERGENCY;

	xil_printf("%s%s", level_strs[level], msgOut);
}
